import cv2
import numpy as np
import os

from . import model as _model
from . import utils as _utils

class Preprocessing(_model.Model):
    '''
    This class provides functionality of the image preprocessing.

    Example::

        preproc = nnio.Preprocessing(
            resize=(224, 224),
            dtype='float32',
            divide_by_255=True,
            means=[0.485, 0.456, 0.406],
            stds=[0.229, 0.224, 0.225],
            batch_dimension=True,
            channels_first=True,
        )

        # Use with numpy image
        image_preprocessed = preproc(image_rgb)

        # Or use to read image from disk
        image_preprocessed = preproc('path/to/image.png')

        # Or use to read image from the web
        image_preprocessed = preproc('http://www.example.com/image.png')

    Object of this type is returned every time you call ``get_preprocessing()`` method of any model from :ref:`nnio.zoo`.
    '''
    def __init__(
        self,
        resize=None,
        dtype=None,
        divide_by_255=None,
        means=None,
        stds=None,
        scales=None,
        imagenet_scaling=False,
        to_gray=None,
        padding=False,
        channels_first=False,
        batch_dimension=False,
        bgr=False,
    ):
        '''
        :parameter resize: ``None`` or ``tuple``.
            (width, height) - the new size of image
        :parameter dtype: ``str`` or ``np.dtype``.
            Data type. By default will use ``uint8``.
        :parameter divide_by_255: ``bool``.
            Divide input image by 255. This is applied before ``means``, ``stds`` and ``scales``.
        :parameter means: ``float`` or iterable or ``None``.
            Substract these values from each channel
        :parameter stds: `float`` or iterable or ``None``.
            Divide each channel by these values
        :parameter scales: ``float`` or iterable or ``None``.
            Multipy each channel by these values
        :parameter imagenet_scaling: apply imagenet scaling.
            It is equivalent to ``divide_by_255=True, means=[0.485, 0.456, 0.406], stds=[0.229, 0.224, 0.225]``.
            If this is specified, arguments ``divide_by_255``, ``means``, ``stds``, ``scales`` must be ``None``.
        :parameter to_gray: if ``int``, then convert rgb image to grayscale with specified number of channels (usually 1 or 3).
        :parameter padding: ``bool``.
            If ``True``, images will be resized with the same aspect ratio
        :parameter channels_first: ``bool``.
            If ``True``, image will be returned in ``[B]CHW`` format.
            If ``False``, ``[B]HWC``.
        :parameter batch_dimension: ``bool``.
            If ``True``, add first dimension of size 1.
        :parameter bgr: ``bool``.
            If ``True``, change channels to BRG order.
            If ``False``, keep the RGB order.
        '''
        self.resize = resize
        self.dtype = dtype or 'uint8'
        self.divide_by_255 = divide_by_255 or False
        if divide_by_255:
            MSG = 'If dividing image by 255, specify float data type'
            assert 'float' in dtype, MSG
        self.means = means
        if stds is not None and scales is not None:
            raise BaseException('Either scales or stds may be specified, not both')
        self.stds = stds
        self.scales = scales
        self.to_gray = to_gray
        assert to_gray is None or isinstance(to_gray, int)
        if to_gray is not None and to_gray <= 0 and channels_first:
            raise BaseException('to_gray <= 0 means that there will be no channel dimension, but found channels_first=True')
        self.padding = padding
        self.channels_first = channels_first
        self.batch_dimension = batch_dimension
        self.bgr = bgr

        if imagenet_scaling:
            if (
                divide_by_255 is not None
                or
                means is not None
                or
                stds is not None
                or
                scales is not None
            ):
                raise BaseException('If imagenet_scaling is True, then arguments divide_by_255, means, stds, scales must be None')
            MSG = 'If imagenet_scaling is True, specify float data type'
            assert dtype is None or 'float' in dtype, MSG
            if dtype is None:
                self.dtype = 'float32'
            self.means = [123.675, 116.28, 103.53]
            self.stds = [58.395, 57.12, 57.375]
        
        # Optimize
        self._scales = None if self.scales is None else np.array(self.scales)[None, None]
        self._means = None if self.means is None else np.array(self.means)[None, None]
        if self.stds is not None:
            self._scales = 1 / np.array(self.stds)[None, None]
        if self.divide_by_255:
            if self._scales is not None:
                self._scales = self._scales / 255
            else:
                self._scales = 1 / 255
            if self._means is not None:
                self._means = self._means * 255

    def forward(self, image, return_original=False):
        '''
        Preprocess the image.

        :parameter image: np.ndarray of type ``uint8`` or ``str``
            RGB image
            If ``str``, it will be concerned as image path.
        :parameter return_original: ``bool``.
            If ``True``, will return tuple of ``(preprocessed_image, original_image)``
        '''
        # Read image
        if isinstance(image, str):
            image = self._read_image(image)
        if return_original:
            orig_image = image.copy()
        if str(image.dtype) != 'uint8':
            raise BaseException('Input image data type for preprocessor must be uint8')

        # Convert colors
        if self.bgr:
            image = image[:, :, ::-1]

        # Resize image
        if self.resize is not None:
            image = self._resize_image(image, self.resize, self.padding)

        # Shift and scale
        if self._means is not None:
            image = image - self._means
        if self._scales is not None:
            image = image * self._scales

        # Convert to grayscale
        if self.to_gray is not None:
            image = image.mean(2, keepdims=self.to_gray > 0)
            if self.to_gray > 1:
                image = image.repeat(self.to_gray, axis=2)

        # Change shape
        if self.channels_first:
            image = image.transpose([2, 0, 1])
        if self.batch_dimension:
            image = image[None]

        # Change datatype
        image = image.astype(self.dtype)

        if return_original:
            return image.copy(), orig_image
        else:
            return image.copy()

    @staticmethod
    def _read_image(path):
        ''' Read image from file or url '''
        # Download image if path is url
        is_url = _utils.is_url(path)
        if is_url:
            path = _utils.file_from_url(path, 'temp', use_cached=False)
        # Read image
        # pylint: disable=no-member
        image = cv2.imread(path)
        # Delete temporary file
        if is_url:
            os.remove(path)
        # Throw exception
        if image is None:
            raise BaseException('Cannot read ' + path)
        # Convert from BGR to RGB
        image = image[:, :, ::-1]
        return image

    @staticmethod
    def _resize_image(image, resize, padding=False):
        ''' Resize image
        
        :parameter image: np.ndarray of type ``uint8`` or ``str``
            RGB image
            If ``str``, it will be concerned as image path.
        :parameter resize: ``None`` or ``tuple``.
            (width, height) - the new size of image
        :parameter padding: ``bool``.
            If ``True``, images will be resized with the same aspect ratio
        :return: np.array. Resized image.
        '''
        if not padding:
            # pylint: disable=no-member
            image = cv2.resize(image, resize)
        else:
            # Resize saving the aspect ratio
            ratio_0 = image.shape[1] / resize[0]
            ratio_1 = image.shape[0] / resize[1]
            ratio = max(ratio_0, ratio_1)
            new_size = (
                int(image.shape[1] / ratio),
                int(image.shape[0] / ratio)
            )
            # pylint: disable=no-member
            image = cv2.resize(image, new_size)
            # Pad with zeros
            padding = np.zeros([resize[1], resize[0], 3],
                               dtype=image.dtype)
            start_0 = (padding.shape[0] - image.shape[0]) // 2
            start_1 = (padding.shape[1] - image.shape[1]) // 2
            padding[
                start_0: start_0 + image.shape[0],
                start_1: start_1 + image.shape[1],
            ] = image
            image = padding.copy()
        return image

    def __str__(self):
        '''
        :return: full description of the ``Preprocessing`` object
        '''
        s = 'nnio.Preprocessing(resize={}, dtype={}, divide_by_255={}, means={}, stds={}, scales={}, to_gray={}, padding={}, channels_first={}, batch_dimension={}, bgr={})'
        s = s.format(
            self.resize,
            self.dtype,
            self.divide_by_255,
            self.means,
            self.stds,
            self.scales,
            self.to_gray,
            self.padding,
            self.channels_first,
            self.batch_dimension,
            self.bgr,
        )
        return s

    def __eq__(self, other):
        '''Compare two ``Preprocessing`` objects. Returns ``True`` only if all preprocessing parameters are the same.'''
        return str(self) == str(other)



from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import onnx as _onnx
from ... import output as _output


class SSDMobileNetV1(_model.Model):
    '''
    SSDMobileNetV1 object detection model trained on COCO dataset.

    Model is taken from the `ONNX Model Zoo <https://github.com/onnx/models>`_.

    Here is the `webcam demo <https://github.com/FastSense/nnio/tree/master/demos>`_ of this model working.
    '''

    URL_MODEL = 'https://github.com/onnx/models/raw/master/vision/object_detection_segmentation/ssd-mobilenetv1/model/ssd_mobilenet_v1_10.onnx'
    URL_LABELS = 'https://github.com/amikelive/coco-labels/raw/master/coco-labels-paper.txt'

    def __init__(self):
        '''
        '''
        super().__init__()

        # Load model
        self.model = _onnx.ONNXModel(self.URL_MODEL)

        # Load labels from text file
        labels_path = _utils.file_from_url(self.URL_LABELS, 'labels')
        self._labels = [
            line.strip()
            for line in open(labels_path)
        ]

    def forward(self, image, return_info=False):
        '''
        :parameter image: np array.
            Input image
        :parameter return_info: bool.
            If ``True``, return inference time.
        :return: list of :class:`nnio.DetectionBox`
        '''
        results = self.model(image, return_info=return_info)
        if return_info:
            results, info = results
        boxes, classes, scores, num_detections = results
        # Parse output
        classes = classes - 1
        out_boxes = []
        for i in range(int(num_detections[0])):
            y_min, x_min, y_max, x_max = boxes[0, i]
            label = self.labels[int(classes[0, i])]
            score = scores[0, i]
            out_boxes.append(
                _output.DetectionBox(x_min, y_min, x_max, y_max, label, score)
            )
        if return_info:
            return out_boxes, info
        else:
            return out_boxes

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(224, 224),
            dtype='uint8',
            batch_dimension=True,
        )

    @property
    def labels(self):
        '''
        :return: list of COCO labels
        '''
        return self._labels



import platform
import time

from . import model as _model
from . import utils as _utils

EDGETPU_SHARED_LIB = {
  'Linux': 'libedgetpu.so.1',
  'Darwin': 'libedgetpu.1.dylib',
  'Windows': 'edgetpu.dll'
}[platform.system()]


class EdgeTPUModel(_model.Model):
    '''
    This class works with tflite models on CPU and with quantized tflite models on Google Coral Edge TPU.

    Using this class requires some libraries to be installed. See :ref:`installation`.
    '''
    def __init__(
        self,
        model_path: str,
        device='CPU'
    ):
        '''
        :parameter model_path: URL or path to the tflite model
        :parameter device: str.
            ``CPU`` by default.
            Set ``TPU`` or ``TPU:0`` to use the first EdgeTPU device.
            Set ``TPU:1`` to use the second EdgeTPU device etc.
        '''
        super().__init__()
        # Download file from internet
        if _utils.is_url(model_path):
            model_path = _utils.file_from_url(model_path, 'models')
        # Create interpreter
        assert device == 'CPU' or device.split(':')[0] == 'TPU' or device[0] == ':'
        self.interpreter = self._make_interpreter(model_path, device)
        self.interpreter.allocate_tensors()

    def forward(self, *inputs, return_info=False):
        assert len(inputs) == self.n_inputs
        start = time.time()
        # Put input tensors into model
        for i in range(self.n_inputs):
            tensor = self._input_tensor(i)
            tensor[:, :, :, :] = inputs[i]
            del tensor
        before_invoke = time.time()
        # Call model
        self.interpreter.invoke()
        after_invoke = time.time()
        # Get results from the model
        results = [self._output_tensor(i) for i in range(self.n_outputs)]
        # Process output a little
        if self.n_outputs == 1:
            results = results[0]
        # Return results
        if return_info:
            info = {
                'assign_time': before_invoke - start,
                'invoke_time': after_invoke - before_invoke,
            }
            return results, info
        else:
            return results

    def get_input_details(self):
        return [
            {
                'name': inp['name'],
                'shape': inp['shape'],
                'dtype': str(inp['dtype']),
            }
            for inp in self.interpreter.get_input_details()
        ]

    def get_output_details(self):
        return [
            {
                'name': inp['name'],
                'shape': inp['shape'],
                'dtype': str(inp['dtype']),
            }
            for inp in self.interpreter.get_output_details()
        ]

    @staticmethod
    def _make_interpreter(model_file, device='CPU'):
        ' Load model and create tflite interpreter '
        try:
            import tflite_runtime.interpreter as tflite
        except ImportError:
            print('''
            Warning: tflite_runtime is not installed.
            
            Please follow these instructions to install driver:
            
            https://coral.ai/docs/m2/get-started/#2a-on-linux
            
            And these instructions to install tflite_runtime:
            
            https://www.tensorflow.org/lite/guide/python
            ''')
            if device == 'CPU':
                print('Trying to use tensorflow version on CPU')
                import tensorflow.lite as tflite
            else:
                raise ImportError
        if device != 'CPU':
            if device == 'TPU':
                device = 'TPU:0'
            return tflite.Interpreter(
                model_path=model_file,
                experimental_delegates=[
                    tflite.load_delegate(
                        EDGETPU_SHARED_LIB,
                        {
                            'device': device.replace('TPU', '')
                        }
                    )
                ])
        else:
            return tflite.Interpreter(
                model_path=model_file)

    def _input_tensor(self, i=0):
        '''
        Returns input tensor view as function returning numpy array
        '''
        tensor_index = self.interpreter.get_input_details()[i]['index']
        return self.interpreter.tensor(tensor_index)()

    def _output_tensor(self, i=0):
        """Returns output tensor view."""
        tensor_index = self.interpreter.get_output_details()[i]['index']
        tensor = self.interpreter.get_tensor(tensor_index)
        return tensor

    @property
    def n_inputs(self):
        ''' number of input tensors '''
        return len(self.interpreter.get_input_details())

    @property
    def n_outputs(self):
        ''' number of output tensors '''
        return len(self.interpreter.get_output_details())




import abc


class Model(abc.ABC):
    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs)

    @abc.abstractmethod
    def forward(self, *args, **kwargs):
        r'''
        This method is called when the model is called.

        :parameter \*inputs: numpy arrays, Inputs to the model
        :parameter return_info: bool, If True, will return inference time
        :return: numpy array or list of numpy arrays.
        '''

    def get_preprocessing(self):
        """
        :return: :class:`nnio.Preprocessing` object.
        """

    def get_input_details(self):
        """
        :return: human-readable model input details.
        """

    def get_output_details(self):
        """
        :return: human-readable model output details.
        """



__version__ = '0.3.1.1'

import nnio.zoo

# Base model class
from .model import Model

# Models for specific backends
from .edgetpu import EdgeTPUModel
from .openvino import OpenVINOModel
from .onnx import ONNXModel
from .pytorch import TorchModel

# Preprocessing class
from .preprocessing import Preprocessing

# Output classes
from .output import DetectionBox



from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import edgetpu as _edgetpu


class MobileNet(_model.Model):
    '''
    MobileNet V2 (or V1) classifier trained on ImageNet

    Model is taken from the `google-coral repo <https://github.com/google-coral/edgetpu/tree/master/test_data>`_
    '''

    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/mobilenet_{}_1.0_224_quant.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/mobilenet_{}_1.0_224_quant_edgetpu.tflite'
    URL_LABELS = 'https://github.com/google-coral/edgetpu/raw/master/test_data/imagenet_labels.txt'

    def __init__(self, device='CPU', version='v2'):
        '''
        :parameter device: str.
            ``CPU`` by default.
            Set ``TPU`` or ``TPU:0`` to use the first EdgeTPU device.
            Set ``TPU:1`` to use the second EdgeTPU device etc.
        :parameter version: str.
            Either ``v1`` or ``v2``.
        '''
        super().__init__()

        # Load model
        if device == 'CPU':
            model_path = self.URL_CPU.format(version)
        else:
            model_path = self.URL_TPU.format(version)
        self.model = _edgetpu.EdgeTPUModel(model_path, device)

        # Load labels from text file
        labels_path = _utils.file_from_url(self.URL_LABELS, 'labels')
        self._labels = [
            ' '.join(line.strip().split()[1:])
            for line in open(labels_path)
        ]

    def forward(self, image, return_scores=False):
        '''
        :parameter image: np array.
            Input image
        :parameter return_scores: bool.
            If ``True``, return class scores.
        :return: ``str``: class label.
        '''
        scores = self.model(image)[0]
        label = self.labels[scores.argmax()]
        if return_scores:
            return label, scores
        else:
            return label

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(224, 224),
            dtype='uint8',
            padding=False,
            batch_dimension=True,
        )

    @property
    def labels(self):
        '''
        :return: list of ImageNet classification labels
        '''
        return self._labels



from setuptools import setup
import codecs
import os.path

def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r', encoding='utf-8') as fp:
        return fp.read()

def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1].strip()
    else:
        raise RuntimeError("Unable to find version string.")


LONG_DESCRIPTION = '''
Please refer to the [project's documentation](https://nnio.readthedocs.io/).

## What is it

**nnio** is a light-weight python package for easily running neural networks.

It supports running models on CPU as well as some of the edge devices:

* [Google USB Accelerator](https://coral.ai/products/accelerator/)
* [Intel Compute Stick](https://www.intel.ru/content/www/ru/ru/products/boards-kits/compute-stick.html)
* Intel integrated GPUs

For each device there exists an own library and a model format. We wrap all those in a single well-defined python package.

Look at this simple example:

```python
import nnio

# Create model and put it on a Google Coral Edge TPU device
model = nnio.EdgeTPUModel(
    model_path='path/to/model_quant_edgetpu.tflite',
    device='TPU',
)
# Create preprocessor
preproc = nnio.Preprocessing(
    resize=(224, 224),
    batch_dimension=True,
)

# Preprocess your numpy image
image = preproc(image_rgb)

# Make prediction
class_scores = model(image)
```

**nnio** was developed for the [Fast Sense X](https://fastsense.readthedocs.io/en/latest/) microcomputer.
It has **six neural accelerators**, which are all supported by nnio:

* 3 x [Google Coral Edge TPU](https://coral.ai/)
* 2 x [Intel VPU](https://www.intel.ru/content/www/ru/ru/products/processors/movidius-vpu/movidius-myriad-x.html)
* an integrated Intel GPU

[More usage examples](https://nnio.readthedocs.io/en/latest/basic_usage.html) can be found in the [documentation](https://nnio.readthedocs.io/).
'''

setup(
    name='nnio',
    version=get_version('nnio/__init__.py'),
    description='Neural network inference on accelerators simplified',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Ruslan Baynazarov',
    author_email='ruslan.baynazarov@fastsense.tech',
    url='https://github.com/FastSense/nnio',
    packages=['nnio', 'nnio.zoo', 'nnio.zoo.edgetpu', 'nnio.zoo.openvino', 'nnio.zoo.onnx'],
    license="MIT",
    install_requires=[
        'numpy',
        'opencv-python',
    ]
)



from ... import utils as _utils
from ... import preprocessing as _preprocessing

from ... import model as _model
from ... import edgetpu as _edgetpu


class DeepLabV3(_model.Model):
    '''
    DeepLabV3 instance segmentation model trained in Pascal VOC dataset.

    Model is taken from the `google-coral repo <https://github.com/google-coral/edgetpu/tree/master/test_data>`_.
    '''

    URL_CPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/deeplabv3_mnv2_dm05_pascal_quant.tflite'
    URL_TPU = 'https://github.com/google-coral/edgetpu/raw/master/test_data/deeplabv3_mnv2_dm05_pascal_quant_edgetpu.tflite'
    URL_LABELS = 'https://github.com/google-coral/edgetpu/raw/master/test_data/pascal_voc_segmentation_labels.txt'

    def __init__(self, device='CPU'):
        '''
        :parameter device: str.
            ``CPU`` by default.
            Set ``TPU`` or ``TPU:0`` to use the first EdgeTPU device.
            Set ``TPU:1`` to use the second EdgeTPU device etc.
        '''
        super().__init__()

        # Load model
        if device == 'CPU':
            model_path = self.URL_CPU
        else:
            model_path = self.URL_TPU
        self.model = _edgetpu.EdgeTPUModel(model_path, device)

        # Load labels from text file
        labels_path = _utils.file_from_url(self.URL_LABELS, 'labels')
        self._labels = []
        for line in open(labels_path):
            if line.strip() != '':
                self.labels.append(line.strip())
            else:
                break

    def forward(self, image):
        '''
        :parameter image: np array.
            Input image
        :return: numpy array.
            Segmentation map of the same size as the input image: ``shape=[batch, 513, 513]``.
            For each pixel gives an integer denoting class.
            Class labels are available through ``.labels`` attribute of this object.
        '''
        segmentation = self.model(image)[0]
        return segmentation

    def get_preprocessing(self):
        return _preprocessing.Preprocessing(
            resize=(513, 513),
            dtype='uint8',
            padding=False,
            batch_dimension=True,
        )

    @property
    def labels(self):
        '''
        :return: list of Pascal VOC labels
        '''
        return self._labels



import time

from . import model as _model
from . import utils as _utils


class TorchModel(_model.Model):
    '''
    This class is used with saved torchscript models.

    For saving model, use `torch.jit.trace` (easier) or `torch.jit.script` (harder).

    Usage example::

        # Create model
        model = nnio.TorchModel('path/to/model.pt')
        # Create preprocessor
        preproc = nnio.Preprocessing(
            resize=(300, 300),
            dtype='uint8',
            batch_dimension=True,
            channels_first=True,
        )

        # Preprocess your numpy image
        image = preproc(image_rgb)

        # Make prediction
        class_scores = model(image)


    Using this class requires torch to be installed. See :ref:`installation`.
    '''
    def __init__(
        self,
        model_path: str,
        device: str='cpu',
    ):
        '''

        :parameter model_path: URL or path to the torchscript model
        :parameter device: Can be either ``cpu`` or ``cuda``.
        '''
        super().__init__()
        self.device = device

        # Download file from the internet
        if _utils.is_url(model_path):
            model_path = _utils.file_from_url(model_path, 'models')

        import torch
        self.torch = torch
        try:
            self.model = torch.jit.load(model_path)
        except:
            self.model = torch.load(model_path)
        self.model.to(device)
        self.model.eval()


    def forward(self, *inputs, return_info=False):
        # Convert inputs to torch tensors
        # pylint: disable=no-member
        inp_torch = [self.torch.tensor(inp, device=self.device) for inp in inputs]

        # Run network and measure time
        start = time.time()
        with self.torch.no_grad():
            outp_torch = self.model(*inp_torch)
        end = time.time()
        results = outp_torch.cpu().numpy()

        # Return results
        if return_info:
            info = {
                'invoke_time': end - start,
            }
            return results, info
        else:
            return results



import time

from . import model as _model
from . import utils as _utils


class OpenVINOModel(_model.Model):
    '''
    This class works with OpenVINO models on CPU, Intel GPU and Intel Movidius Myriad.

    Using this class requires some libraries to be installed. See :ref:`installation`.
    '''
    def __init__(
        self,
        model_bin: str,
        model_xml: str,
        device='CPU',
    ):
        '''
        :parameter model_bin: URL or path to the openvino binary model file
        :parameter model_xml: URL or path to the openvino xml model file
        :parameter device: str.
            Choose Intel device:
            ``CPU``, ``GPU``, ``MYRIAD``
            If there are multiple devices in your system, you can use indeces:
            ``MYRIAD:0`` but it is not recommended since Intel automatically chooses a free device.
        '''
        super().__init__()

        # Download files from internet
        if _utils.is_url(model_bin):
            model_bin = _utils.file_from_url(model_bin, 'models')
        if _utils.is_url(model_xml):
            model_xml = _utils.file_from_url(model_xml, 'models')

        # Create interpreter
        self.ie, self.net, self.device = self._make_interpreter(model_xml, model_bin, device)

    def forward(self, inputs, return_info=False):
        r'''
        :parameter inputs: numpy array, input to the model
        :parameter return_info: bool, If True, will return inference time
        :return: numpy array or list of numpy arrays.
        '''
        # Find name of the input to the model
        input_name = list(self.net.input_info.keys())[0]
        # Call model
        start = time.time()
        out = self.net.infer({input_name: inputs})
        end = time.time()
        # Process output a little
        if len(out.keys()) == 1:
            out = out[list(out.keys())[0]]
        # Measure temperature
        temperature = None
        if self.device.startswith('MYRIAD.'):
            temperature = self.ie.get_metric(metric_name="DEVICE_THERMAL", device_name=self.device)
            if _utils.LOG_TEMPERATURE:
                _utils.log_temperature(self.device, temperature)
        # Return results
        if return_info:
            info = {
                'invoke_time': end - start,
            }
            if temperature is not None:
                info['temperature'] = temperature
            return out, info
        else:
            return out

    @staticmethod
    def _make_interpreter(model_xml, model_bin, device):
        'Load model and create openvino interpreter'
        try:
            from openvino.inference_engine import IECore
        except ImportError:
            print('''
            Warning: openvino is not installed.
            
            Please install openvino or use openvino docker container.
            ''')
            raise ImportError
        ie = IECore()
        # Get list of MYRIAD devices
        if device.startswith('MYRIAD:'):
            myriads = [
                dev for dev in ie.available_devices
                if 'MYRIAD' in dev
            ]
            _, idx = device.split(':')
            idx = int(idx)
            if len(myriads) <= idx:
                raise BaseException('Cannot find out which device is {}\nAvailable devices: {}'.format(device, ie.available_devices))
            device = myriads[idx]
        # Load model on device
        net = ie.read_network(model_xml, model_bin)
        print('Loading model to:', device)
        net = ie.load_network(net, device)
        return ie, net, device
